from __future__ import absolute_import
import os
import warnings
from flask import Flask, current_app
import six
from werkzeug.utils import import_string

# Shared objects
from . import db, csrf, requests_session
# Internal modules/packages
from . import util


def create_app(config=None, **kwargs):
    # Internal modules/packages
    from . import login, divisions, requests, views, json
    app = Flask('evesrp', **kwargs)
    app.request_class = util.AcceptRequest
    app.config.from_object('evesrp.default_config')
    if config is not None:
        app.config.from_pyfile(config)
    if app.config['SECRET_KEY'] is None and 'SECRET_KEY' in os.environ:
        app.config['SECRET_KEY'] = os.environ['SECRET_KEY']

    # Register SQLAlchemy monitoring before the DB is connected
    app.before_request(sqlalchemy_before)

    db.init_app(app)

    login.login_manager.init_app(app)

    before_csrf = list(app.before_request_funcs[None])
    csrf.init_app(app)
    # Remove the context processor that checks CSRF values. All it is used for
    # is the template function.
    app.before_request_funcs[None] = before_csrf

    app.add_url_rule(rule=u'/', view_func=views.index)
    for error_code in (400, 403, 404, 500):
        app.register_error_handler(error_code, views.error_page)
    app.register_blueprint(divisions.blueprint, url_prefix='/division')
    app.register_blueprint(login.blueprint)
    app.register_blueprint(requests.blueprint, url_prefix='/request')
    app.register_blueprint(views.api.api, url_prefix='/api')
    app.register_blueprint(views.api.filters, url_prefix='/api/filter')

    app.add_template_global(requests.request_count)

    app.json_encoder = json.SRPEncoder

    _config_authmethods(app)
    _config_requests_session(app)
    _config_killmails(app)
    _config_url_converters(app)

    # Configure the Jinja context
    # Inject variables into the context
    from .permission import PermissionType
    @app.context_processor
    def inject_enums():
        return {
            'ActionType': requests.ActionType,
            'PermissionType': PermissionType,
            'app_version': __version__,
            'site_name': app.config['SRP_SITE_NAME'],
            'url_for_page': requests.url_for_page,
        }
    # Auto-trim whitespace
    app.jinja_env.trim_blocks = True
    app.jinja_env.lstrip_blocks = True

    return app


# SQLAlchemy performance logging
def sqlalchemy_before():
    if util.DB_STATS.total_queries > 0:
        current_app.logger.debug(u"{} queries in {} ms.".format(
                DB_STATS.total_queries,
                round(DB_STATS.total_time * 1000, 3)))
    util.DB_STATS.clear()
    g.DB_STATS = util.DB_STATS


# Utility function for creating instances from dicts
def _instance_from_dict(instance_descriptor):
    type_name = instance_descriptor.pop('type')
    Type = import_string(type_name)
    return Type(**instance_descriptor)


# Utility function for raising config deprecation warnings
def _deprecated_object_instance(key, value):
    warnings.warn(u"Non-basic data types in configuration values are deprecated"
                 u"({}: {})".format(key, value), DeprecationWarning,
                 stacklevel=2)


# Auth setup
def _config_authmethods(app):
    auth_methods = []
    # Once the deprecated config value support is removed, this can be
    # rewritten as a dict comprehension
    for method in app.config['SRP_AUTH_METHODS']:
        if isinstance(method, dict):
            auth_methods.append(_instance_from_dict(method))
        elif isinstance(method, login.AuthMethod):
            _deprecated_object_instance('SRP_AUTH_METHODS', method)
            auth_methods.append(method)
    app.auth_methods = auth_methods


# Request detail URL setup
def _config_url_converters(app):
    url_transformers = {}
    for config_key, config_value in app.config.items():
        # Skip null config values
        if config_value is None:
            continue
        # Look for config keys in the format 'SRP_*_URL_TRANSFORMERS'
        if not config_key.endswith('_URL_TRANSFORMERS')\
                or not config_key.startswith('SRP_'):
            continue
        attribute = config_key.replace('SRP_', '', 1)
        attribute = attribute.replace('_URL_TRANSFORMERS', '', 1)
        attribute = attribute.lower()
        # Create Transformer instances for this attribute
        transformers = {}
        for transformer_config in config_value:
            if isinstance(transformer_config, tuple):
                # Special case for Transformers: A 2-tuple in the form:
                # (u'Transformer Name',
                #     'http://example.com/transformer/slug/{}')
                transformer = divisions.Transformer(*transformer_config)
            elif isinstance(transformer_config, dict):
                # Standard instance dictionary format
                # Provide a default type value
                transformer_config.setdefault('type',
                        'evesrp.divisions.Transformer')
                transformer = _instance_from_dict(transformer_config)
            elif isinstance(transformer_config, divisions.Transformer):
                # DEPRECATED: raw Transformer instance
                _deprecated_object_instance(config_key, transformer_config)
                transformer = transformer_config
            transformers[transformer.name] = transformer
        url_transformers[attribute] = transformers
    app.url_transformers = url_transformers


# Requests session setup
def _config_requests_session(app):
    try:
        ua_string = app.config['SRP_USER_AGENT_STRING']
    except KeyError as outer_exc:
        try:
            ua_string = 'EVE-SRP/{version} ({email})'.format(
                    email=app.config['SRP_USER_AGENT_EMAIL'],
                    version=__version__)
        except KeyError as inner_exc:
            raise inner_exc
    requests_session.headers.update({'User-Agent': ua_string})
    app.user_agent = ua_string


# Killmail verification
def _config_killmails(app):
    killmail_sources = []
    # For now, use a loop with checks. After removing the depecated config
    # method it can be rewritten as a list comprehension
    for source in app.config['SRP_KILLMAIL_SOURCES']:
        if isinstance(source, six.string_types):
            killmail_sources.append(import_string(source))
        elif isinstance(source, type):
            _deprecated_object_instance('SRP_KILLMAIL_SOURCES', source)
            killmail_sources.append(source)
    app.killmail_sources = killmail_sources



#!/usr/bin.env python

from __future__ import absolute_import
from __future__ import print_function
import os
import os.path
import argparse
from itertools import cycle
import json
from decimal import Decimal
import time
import datetime as dt
import flask
from flask.ext import script
from flask.ext.migrate import Migrate, MigrateCommand, _get_config, stamp
from alembic.migration import MigrationContext
from alembic.script import ScriptDirectory
import six
from .. import create_app, db, migrate, models, auth, killmail
from .datetime import utc


if six.PY3:
    unicode = str


manager = script.Manager(create_app)


class AbsolutePathAction(argparse.Action):
    """Custom argparse.Action that transforms path strings into absolute paths.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        current_absolute = os.path.abspath(os.getcwd())
        if isinstance(values, str):
            new_values = os.path.join(current_absolute, values)
        else:
            new_values = []
            for value in values:
                real_path = os.path.join(current_absolute, values)
                new_values.append(real_path)
        setattr(namespace, self.dest, new_values)


# Monkeypatch Flask-Script to consider my custom path action 'safe'
safe_actions = list(script.safe_actions)
safe_actions.append(AbsolutePathAction)
script.safe_actions = safe_actions


# Crufty workaround to get Flask-Migrate working with app factories
class MigrateManager(script.Manager):
    def __init__(self, old_command, directory='migrations'):
        super(MigrateManager, self).__init__()
        self.directory = directory
        for attr in ('app', '_commands', '_options', 'usage', 'help',
                'description', 'disable_argcomplete', 'with_default_commands',
                'parent'):
            setattr(self, attr, getattr(old_command, attr))

    def __call__(self, app=None, directory='migrations', **kwargs):
        if app is None:
            app = self.app
            if app is None:
                raise Exception(u"No app specified")

        db = kwargs.pop('db', None)

        if not isinstance(app, flask.Flask):
            app = app(**kwargs)

        # Last ditch effort to get a database handle
        if db is None:
            if 'sqlalchemy' in app.extensions:
                db = app.extensions['sqlalchemy'].db
            else:
                raise Exception(u"No database defined for app.")

        Migrate(app, db, self.directory)
        return app


migrate_path = os.path.dirname(migrate.__file__)
migrate_path = os.path.abspath(migrate_path)
migrate_manager = MigrateManager(MigrateCommand, migrate_path)
manager.add_command('db', migrate_manager)


manager.add_option('-c', '--config', dest='config', required=True,
        action=AbsolutePathAction)


@migrate_manager.command
def create(force=False):
    """Create tables if the database has not been configured yet."""
    # Fail if there's an alembic version set
    engine = db.get_engine(flask.current_app)
    conn = engine.connect()
    context = MigrationContext.configure(conn)
    current_rev = context.get_current_revision()
    alembic_config = _get_config(directory=migrate_path)
    script = ScriptDirectory.from_config(alembic_config)
    latest_rev = script.get_current_head()
    if current_rev == latest_rev and not force:
        print(u"You need to run 'evesrp -c config.py db migrate' to "
              u"migrate to the latest database schema.")
    else:
        db.create_all()
        if current_rev is None:
            stamp()


@manager.shell
def shell_context():
    ctx = dict(
        app=flask.current_app,
        db=db)
    for cls in ('Request', 'Action', 'ActionType', 'Modifier',
                'AbsoluteModifier', 'RelativeModifier'):
        ctx[cls] = getattr(models, cls)
    for cls in ('Entity', 'User', 'Group', 'Note', 'APIKey', 'Permission',
            'Division', 'Pilot'):
        ctx[cls] = getattr(auth.models, cls)
    ctx['PermissionType'] = auth.PermissionType
    return ctx


class PopulatedKillmail(killmail.Killmail, killmail.RequestsSessionMixin,
        killmail.ShipNameMixin, killmail.LocationMixin):
    pass


class Populate(script.Command):
    """Populate the database with data given from a list of losses form zKB."""

    option_list = (
        script.Option('--file', '-f', dest='kill_file', required=True),
        script.Option('--users', '-u', dest='num_users', default=5, type=int),
        script.Option('--divisions', '-d', dest='num_divisions', default=3,
                type=int),
    )

    def run(self, kill_file, num_users, num_divisions, **kwargs):
        # Set up users, divisions and permissions
        users = []
        for user_num, authmethod in zip(range(num_users),
                cycle(flask.current_app.config['SRP_AUTH_METHODS'])):
            user_name = u'User {}'.format(user_num)
            user = auth.models.User.query.filter_by(name=user_name).first()
            if user is None:
                user = auth.models.User(user_name, authmethod.name)
                db.session.add(user)
            users.append(user)
        divisions = []
        for division_num in range(num_divisions):
            division_name = u'Division {}'.format(division_num)
            division = auth.models.Division.query.filter_by(
                    name=division_name).first()
            if division is None:
                division = auth.models.Division(division_name)
                db.session.add(division)
            divisions.append(division)
        db.session.add_all(divisions)
        for user in users:
            for division in divisions:
                perm = auth.models.Permission.query.filter_by(division=division,
                        permission=auth.PermissionType.submit,
                        entity=user).first()
                if perm is None:
                    auth.models.Permission(division,
                            auth.PermissionType.submit, user)
        db.session.commit()
        # load and start processing killmails
        with open(kill_file, 'r') as f:
            kills = json.load(f)
        for user, division, kill_info in zip(cycle(users), cycle(divisions), kills):
            victim = kill_info[u'victim']
            # Skip corp-level kills (towers, pocos, etc)
            if victim['characterID'] == '0':
                continue
            # make sure a Pilot exists for this killmail
            pilot = auth.models.Pilot.query.get(victim['characterID'])
            if pilot is None:
                pilot = auth.models.Pilot(user, victim['characterName'],
                        int(victim['characterID']))
            db.session.commit()
            pilot_user = pilot.user
            # create a Killmail
            args = dict(
                    kill_id=int(kill_info[u'killID']),
                    pilot_id=int(victim[u'characterID']),
                    pilot=victim[u'characterName'],
                    corp_id=int(victim[u'corporationID']),
                    corp=victim[u'corporationName'],
                    ship_id=int(victim[u'shipTypeID']),
                    system_id=int(kill_info[u'solarSystemID']),
                    verified=True)
            if victim[u'allianceID'] != '0':
                args['alliance_id'] = int(victim[u'allianceID'])
                args['alliance'] = victim[u'allianceName']
            time_struct = time.strptime(kill_info[u'killTime'], '%Y-%m-%d %H:%M:%S')
            args['timestamp'] = dt.datetime(*(time_struct[0:6]), tzinfo=utc)
            args['url'] = u'https://zkillboard.com/kill/{}'.format(
                    args['kill_id'])
            try:
                args['value'] = Decimal(kill_info[u'zkb'][u'totalValue'])
            except KeyError:
                args['value'] = Decimal(0)
            killmail = PopulatedKillmail(**args)
            try:
                killmail.ship
            except KeyError:
                continue
            # Create a request for this killmail
            models.Request(pilot_user, unicode(killmail), division, killmail)
            db.session.commit()


manager.add_command('populate', Populate())


def main():
    manager.run()


if __name__ == '__main__':
    main()



from __future__ import absolute_import
from __future__ import unicode_literals
from .killmail import CRESTMail, ZKillmail

CSRF_ENABLE = True

# This really needs to be set to something
SRP_AUTH_METHODS = []

# Default database is an in-memory SQLite DB.
SQLALCHEMY_DATABASE_URI = 'sqlite://'

# CREST killmails don't have any external dependencies
SRP_KILLMAIL_SOURCES = [ZKillmail, CRESTMail]

SRP_SITE_NAME = 'EVE-SRP'

SRP_DETAILS_DESCRIPTION = 'Supporting details about your loss.'

SRP_SKIP_VALIDATION = False



from __future__ import absolute_import
from flask import flash, redirect, current_app, url_for, request
from flask.ext.login import current_user
from flask.ext.oauthlib.client import OAuthException
from flask.ext.wtf import csrf
from sqlalchemy.orm.exc import NoResultFound

from .. import db, oauth
from ..util import ensure_unicode
from . import AuthMethod
from .models import User, Group, Pilot


def token_getter():
    """Simple function to retrieve an access token from the current logged in
    user.
    """
    return {'access_token': current_user.token}


class OAuthMethod(AuthMethod):

    def __init__(self, **kwargs):
        """Abstract :py:class:`~.AuthMethod` for OAuth-based login methods.

        Implementing classes need to implement :py:meth:`get_user`\,
        :py:meth:`get_pilots`\, and :py:meth:`get_groups`\.

        In addition to the keyword arguments from :py:class:`~.AuthMethod`,
        this initializer accepts the following arguments that will be used in
        the creation of the :py:attr:`OAuthMethod.oauth` object (See the
        documentation for :py:class:`~flask_oauthlib.client.OAuthRemoteApp` for
        more details):

        * ``base_url``
        * ``request_token_url``
        * ``access_token_url``
        * ``authorize_url``
        * ``consumer_key``
        * ``consumer_secret``
        * ``request_token_params``
        * ``access_token_params``
        * ``access_token_method``
        * ``content_type``
        * ``app_key``
        * ``encoding``

        As a convenience, the ``key`` and ``secret`` keyword arguments will be
        treated as ``consumer_key`` and ``consumer_secret`` respectively. The
        ``name`` argument is used for both :py:class:`~.AuthMethod` and for
        :py:class:`~flask_oauthlib.client.OAuthRemoteApp`.

        Subclasses for providers that may be used by more than one entity are
        encouraged to provide their own defaults for the above arguments.

        The redirect URL for derived classes is based off of the
        :py:attr:`~.AuthMethod.safe_name` of the implementing
        :py:class:`~.AuthMethod`, specifically the URL for
        :py:meth:`~.AuthMethod.view`. For example, the default redirect URL for
        :py:class:`~.TestOAuth` is similar to
        ``https://example.com/login/test_oauth/`` (Note the trailing slash, it
        is significant).
        """
        # Allow using 'secret' and 'key' instead of 'consumer_[secret|key]'
        if 'key' in kwargs:
            kwargs['consumer_key'] = kwargs.pop('key')
        if 'secret' in kwargs:
            kwargs['consumer_secret'] = kwargs.pop('secret')
        # Remove OAuth arguments and create an arguments dictionary for them
        oauth_kwargs = {}
        for kwarg in ('oauth', 'base_url', 'request_token_url',
                      'access_token_url', 'authorize_url', 'consumer_key',
                      'consumer_secret', 'request_token_params',
                      'access_token_params', 'access_token_method',
                      'content_type', 'app_key', 'encoding'):
            if kwarg in kwargs:
                oauth_kwargs[kwarg] = kwargs.pop(kwarg)
        if 'name' not in kwargs:
            self.name = 'OAuth'
        else:
            self.name = kwargs['name']
        oauth_kwargs['name'] = self.name
        self.oauth = oauth.remote_app(**oauth_kwargs)
        self.oauth.tokengetter(token_getter)
        self.scope = kwargs.pop('scope', None)
        super(OAuthMethod, self).__init__(**kwargs)

    def login(self, form):
        # CSRF token valid for 5 minutes
        csrf_token = csrf.generate_csrf(time_limit=300)
        return self.oauth.authorize(callback=url_for('login.auth_method_login',
                                    auth_method=self.safe_name,
                                    _external=True), state=csrf_token)

    def view(self):
        """Handle creating and/or logging in the user and updating their
        :py:class:`~.Pilot`\s and :py:class:`~.Group`\s.
        """
        if hasattr(self.oauth, 'authorized_response'):
            resp = self.oauth.authorized_response()
        else:
            resp = self.oauth.handle_response()
        # Check that the response was successful
        # Yeah, an exception as a return value. I don't know either.
        if isinstance(resp, OAuthException):
            flash(u"Login failed: {} ({})".format(resp.type, resp.message),
                    u'error')
            return redirect(url_for('login.login'))
        # Handle other kinds of errors
        elif resp is None:
            reason = ensure_unicode(request.args.get('error',
                    u'Unknown error'))
            if reason == u'access_denied':
                reason = u'Access denied'
            flash(u"Login failed: {}".format(reason), u'error')
            return redirect(url_for('login.login'))
        # Check CSRF token
        csrf_token = request.args['state']
        if not csrf.validate_csrf(csrf_token, time_limit=True):
            flash(u"CSRF validation failed. Please try logging in again.",
                    u'error')
            return redirect(url_for('login.login'))
        token = {'access_token': resp['access_token']}
        # Get the User object for this user, creating one if needed
        user = self.get_user(token)
        if user is not None:
            # Apply site-wide admin flag
            user.admin = self.is_admin(user)
            # Login the user, so current_user will work
            self.login_user(user)
        else:
            flash(u"Login failed.", u'error')
            return redirect(url_for('login.login'))
        # Add new Pilots
        current_pilots = self.get_pilots(token)
        for pilot in current_pilots:
            pilot.user = user
        # Remove old pilots
        user_pilots = set(user.pilots)
        for pilot in user_pilots:
            if pilot not in current_pilots:
                pilot.user = None
        db.session.commit()
        # Add new groups
        current_groups = self.get_groups(token)
        for group in current_groups:
            user.groups.add(group)
        # Remove old groups
        user_groups = set(user.groups)
        for group in user_groups:
            if group not in current_groups and group in user.groups:
                user.groups.remove(group)
        # Save all changes
        db.session.commit()
        return redirect(url_for('index'))

    def get_user(self, token):
        """Returns the :py:class:`~.OAuthUser` instance for the given token.

        This method is to be implemented by subclasses of
        :py:class:`OAuthMethod` to use whatever APIs they have access to to get
        the user account given an access token.

        :param token: The access token used for communicating with whatever API
            you're using.
        :type token: typically a :py:class:`dict` with the `access_token` key's
            value set to the token string.
        :rtype: :py:class:`OAuthUser`
        """
        raise NotImplementedError

    def is_admin(self, user):
        """Returns wether this user should be treated as a site-wide
        administrator.

        The default implementation checks if the user's name is contained
        within the list of administrators supplied as an argument to
        :py:class:`OAuthMethod`\.

        :param user: The user to check.
        :type user: :py:class:`~.OAuthUser`
        :rtype: bool
        """
        return user.name in self.admins

    def get_pilots(self, token):
        """Return a :py:class:`list` of :py:class:`~.Pilot`\s for the given
        token.

        Like :py:meth:`get_user`\, this method is to be implemented by
        :py:class:`OAuthMethod` subclasses to return a list of
        :py:class:`~.Pilot`\s associated with the account for the given access
        token.

        :param token: The access token used for communicating with whatever API
            you're using.
        :type token: typically a :py:class:`dict` with the `access_token` key's
            value set to the token string.
        :rtype: :py:class:`list` of :py:class:`~.Pilot`\s.
        """
        raise NotImplementedError

    def get_groups(self, token):
        """Returns a :py:class:`list` of :py:class:`~.Group`\s for the given
        token.

        Like :py:meth:`get_user` and :py:meth:`get_pilots`\, this method is to
        be implemented by :py:class:`OAuthMethod` subclasses to return a list
        of :py:class:`~.Group`\s associated with the account for the given
        access token.

        :param token: The access token used for communicating with whatever API
            you're using.
        :type token: typically a :py:class:`dict` with the `access_token` key's
            value set to the token string.
        :rtype: :py:class:`list` of :py:class:`~.Group`\s.
        """
        raise NotImplementedError


class OAuthUser(User):
    id = db.Column(db.Integer, db.ForeignKey(User.id), primary_key=True)
    token = db.Column(db.String(100, convert_unicode=True))



from __future__ import absolute_import

import datetime as dt
from sqlalchemy import types


class UTC(dt.tzinfo):

    def utcoffset(self, obj):
        return dt.timedelta()

    def dst(self, obj):
        return dt.timedelta()

    def tzname(self, obj):
        return u'UTC+00:00'


utc = UTC()


class DateTime(types.TypeDecorator):

    impl = types.DateTime

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        if dialect.name == 'mysql':
            return dt.datetime(*(value.utctimetuple()[0:6]))
        return value
